#!/usr/bin/perl -w

use strict;
use IO::Socket::INET;
use IO::Select;
use IO::Handle;
use Data::Dumper;
use Getopt::Long;

# Setup connection information with shell arguments
my $host;
my $addr = 'localhost';
my $uri  = '_check.php';
my $port = 80;
my $tmout = 3;
my $agent = "AnNyung WEB Checker";
my $makeBody = 0;
my $verbose = 0;
my $initscr = "/etc/init.d/httpd";
my $logfile = "/var/log/httpd/httpwatch.log";
my $test = 0;

my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime;
my $status;
my $bodylen = 0;

######################### SUB FUNCTION ######################### {{{
sub trim { # {{{
	my $s = shift;
	$s =~ s/^\s+//g;
	$s =~ s/\s+$//g;
	return $s;
} # }}}

sub rtrim { # {{{
	my $s = shift;
	$s =~ s/(\r)?\n$//;
	return $s;
} # }}}

sub help { # {{{
	print <<EOL;
USAGE: httpWatch [OPTION]
Options:
    -a|--addr=[STRING]        Target Address. [default: localhost]
    -h|--host=[STRING]        HTTP/1.1 Host header
    -u|--uri=[STRING]         HTTP URI for check. [default: _check.php]
                              Removed first slash character.
    -p|--port=[INTEGER]       Port number [default: 80]
    -t|--timeout=[INTEGER]    Connect/Read/Write timeout [default: 3]
    -c|--cagent=[STRING]      Another UserAgent name
    -l|--logfile=[STRING]     Path of log file
    -v|--verbose              Verbose mode
EOL
	exit (1);
} # }}}

sub dprint { # {{{
	return if ( ! $verbose );

	my $format = shift;

	printf $format, @_ if ( $format );
} # }}}

sub restart_daemon { # {{{
	my $rcode = shift;
	my $format = shift;
	my $r = `nice -n -10 $initscr restart >& /dev/null` if ( ! $test );
	$verbose = 1 if ( $test );
	dprint $format, @_;

	$status->{'code'} = -100 if ( ! $status->{'code'} );

	if ( $verbose ) {
		printf "[%04d/%02d/%02d %02d:%02d:%02d] \"http://%s/%s\" \"Host: %s\" Status: %s, Length: %d\n",
				$year + 1900, $mon + 1, $mday, $hour, $min, $sec,
				$addr, $uri, $host, $rcode, $bodylen;
	}
	my $lp = open (LOGF, ">>$logfile");
	if ( $lp ) {
		printf LOGF "[%04d/%02d/%02d %02d:%02d:%02d] \"http://%s/%s\" \"Host: %s\" Status: %s, Length: %d\n",
				$year + 1900, $mon + 1, $mday, $hour, $min, $sec,
				$addr, $uri, $host, $rcode, $bodylen;
		close (LOGF);
	} else {
		printf "Can't open log file(%s) : %s\n", $logfile, $!;
	}

	exit (1);
} # }}}
######################### SUB FUNCTION ######################### }}}

my $opt = GetOptions (
	"addr=s"    => \$addr,
	"host=s"    => \$host,
	"uri=s"     => \$uri,
	"port=i"    => \$port,
	"timeout=i" => \$tmout,
	"cagent=s"   => \$agent,
	"verbose+"  => \$verbose,
	"logfile"   => \$logfile
);

$host = $addr if ( ! $host);

if ( $verbose ) {
	print <<EOL;
----------------------------------------
ADDR    : $addr
HOST    : $host
URI     : /$uri
PORT    : $port
TIMEOUT : $tmout
AGENT   : $agent
----------------------------------------
EOL
}

help () if ( ! $opt );

dprint "* Socket Connection\n";
my $sock = new IO::Socket::INET->new ( # {{{
	PeerAddr => $addr,
	PeerPort => $port,
	Proto    => 'tcp',
	Reuse    => 1,
	Type     => SOCK_STREAM,
	Timeout  => $tmout,
); # }}}

restart_daemon -100, "Could not create socket: %s\n", $! unless $sock;

dprint "* Set socket timeout\n";
$sock->timeout($tmout);
$sock->autoflush (1);

# Sending Query Template
my $msg = <<EOL;
GET /$uri HTTP/1.1\r
Host: $host\r
Connection: close\r
User-Agent: $agent\r
\r
EOL

# use select for timeout
dprint "* Set select for timeout\n";
my $select = IO::Select->new ($sock) or die "IO::Select $!\n";

dprint "* Waiting write event\n";
my @sbuf = $select->can_write ($tmout);
restart_daemon -101, "IO::Select write timeout: %s\n", $! if ( $#sbuf lt 0 );
$sock->send ($msg);

my $hd = IO::Handle->new_from_fd ($sock, "r");

my $bodystart = 0;
my $headers;
my @header;
my $buf;
my $buflen;

my $chunkedlen = 0;
my $chunked = 0;
my $ischunked = 0;

my $body;

dprint "* Waiting read event\n";
while ( 1 ) {
	#dprint "* Waiting read event\n";
	@sbuf = $select->can_read ($tmout);
	if ( $#sbuf lt 0 ) {
		restart_daemon -102, "IO::Select read timeout: %s\n", $!;
		last;
	}

	# Get line buffer
	$msg = $hd->getline () or last;

	# Header parsing {{{
	# ----------------------------------------------------------------------------
	if ( $bodystart == 0 ) {
		if ( ! $status ) {
			if ( $msg =~ /^HTTP\/1\.[0-9] ([0-9]{3}) (.*)/ ) {
				$status->{code} = $1;
				$status->{desc} = trim ($2);

				restart_daemon $status->{code}, "Return code is %d\n", $status->{code} if ( $status->{code} != 200 );
				next;
			} else {
				die "$host:$port is not HTTP service\n";
			}
		}

		# header last
		if ( $msg eq "\r\n" ) {
			$bodystart++;
			next;
		}

		@header = split /:[\s]*/, $msg;
		$headers->{$header[0]} = trim ($header[1]);

		if ( defined ($headers->{'Transfer-Encoding'}) && $headers->{'Transfer-Encoding'} eq 'chunked' ) {
			$ischunked = 1;
		}
		
		next;
	}
	# ---------------------------------------------------------------------------- }}}

	# Chunked Body
	if ( $ischunked ) { # {{{

		$buf = rtrim ($msg);
		$buflen = length ($buf);

		# check chunk
		if ( ($buflen > 0 && $buflen < 5) && $buf =~ /^[0-9a-f]{1,4}$/ ) {
			$chunked = hex ($buf);
			$chunkedlen += $chunked;

			if ( $bodylen > 0 && $chunked > 0 ) {
				$bodylen -= 2;
				$body =~ s/(\r)?\n$//g if ( $makeBody );
			}
			next;
		}

		next if ( $msg eq "\r\n" && $bodylen == $chunkedlen );

		if ( $chunked > 0 ) {
			$bodylen += length ($msg);
			$body .= $msg if ( $makeBody );
		}

		next;
	} # }}}

	# Static Body
	$body .= $msg if ( $makeBody );
	$bodylen += length ($msg);
}

$headers->{'Content-Length'} = $chunkedlen if ( $ischunked );

if ( $headers->{'Content-Length'} != $bodylen ) {
	restart_daemon -103, "Wrong length %d : %d\n", $headers->{'Content-Length'}, $bodylen;
}

if ( $verbose ) {
	print Dumper($headers);
	print Dumper($status);
	if ( $ischunked ) {
		printf "* Chunked   Length : %d\n", $headers->{'Content-Length'};
	} else {
		printf "* Content   Length : %d\n", $headers->{'Content-Length'};
	}
	printf "* Completed Length : %d\n", $bodylen;
	print length ($body) if ( $makeBody );
}

# Writing Log {{{
my $lp = open (LOGF, ">>$logfile");
if ( $lp ) {
	printf LOGF "[%04d/%02d/%02d %02d:%02d:%02d] http://%s/%s \"Host: %s\" Status: %s, Length: %d\n",
		$year + 1900, $mon + 1, $mday, $hour, $min, $sec,
		$addr, $uri, $host, $status->{'code'}, $bodylen;
	close (LOGF);
} else {
	printf "Can't open log file(%s) : %s\n", $logfile, $!;
}
# }}}

$hd->close ();
close $sock;

1;
